/*****************************************************************************
 **
 ** Licensed Material - Property of IBM
 **
 ** 5724-D04
 **
 ** (C) Copyright IBM Corp. 2000 - 2004  All Rights Reserved
 **
 ** US Government Users Restricted Rights - Use, duplication or
 ** disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 **
 ***************************************************************************/
#define ALLOW_ACCESS_TO_INTERNALS_OF_LISTS
#define ALLOW_ACCESS_TO_INTERNALS_OF_FORMS
#define ALLOW_ACCESS_TO_INTERNALS_OF_WINDOWS
#define ALLOW_ACCESS_TO_INTERNALS_OF_CONTROLS
#define ALLOW_ACCESS_TO_INTERNALS_OF_FIELDS

#include <PalmOS.h>
#include <PalmCompatibility.h>

#include "MAB_Print.h"
#include "MAB_Print_Messages.h"
#include "MAB_PalmPrint.h"

#define CHARSPERROW 50
#define TOLERANCE 3

/*
* Private datastructures and functions
*/
static MAB_Print_ListDataHandlerPtr gFunc = NULL;
UInt16 gAlertID = 900;

typedef struct NodeType{
	UInt16 controlID;
	FormPtr pForm;
	UInt32 nMaxLines;
	UInt32 nCurrentLine;
	char *data;
	struct NodeType *next;
}Node;


typedef struct LineType{
	UInt32 nLineNo;
	Node *first;
	UInt32 nElements;
	struct LineType *nextLine;
}Line;

Line *gFirstLine;

#if __MULTI_SEG_APP__
static void getNextLineFromField(Node *thisNode, char **szText) __attribute__ ((section ("printfns")));
static void getNextLineFromLabel(Node *thisNode, char **szText) __attribute__ ((section ("printfns")));
static void getNextLineFromList(Node *thisNode, char **szText) __attribute__ ((section ("printfns")));
static void getNextLineFromNode(Node *thisNode, char **szText) __attribute__ ((section ("printfns")));
static Boolean doesNextLineExist(Node *thisItem) __attribute__ ((section ("printfns")));
static void initField(Node *thisNode) __attribute__ ((section ("printfns")));
static void initLabel(Node *thisNode) __attribute__ ((section ("printfns")));
static void initList(Node *thisNode) __attribute__ ((section ("printfns")));
static Node* initItemNode(FormPtr pForm, UInt32 controlID) __attribute__ ((section ("printfns")));
static UInt32 getCharsVisible(FontID font,UInt16 width) __attribute__ ((section ("printfns")));
static void getLineFromBuffer(UInt32 nLine, UInt32 visibleChars, CharPtr szBuffer, char **szText) __attribute__ ((section ("printfns")));
static void getLineFromList(FormPtr pForm,UInt16 listID, UInt32 lineNo, CharPtr szListData, char **szLine) __attribute__ ((section ("printfns")));
static void createLines(FormPtr pForm, UInt16 ControlsOnForm[], UInt16 nSize) __attribute__ ((section ("printfns")));
static Line* initLine(UInt32 nLineNo) __attribute__ ((section ("printfns")));
static void sortControls(FormPtr pForm, UInt16 **Controls, UInt16 nSize) __attribute__ ((section ("printfns")));
static void createPrintableForm(FormPtr pForm, UInt16 ControlsOnForm[], UInt16 nSize) __attribute__ ((section ("printfns")));
static UInt32 getY(FormPtr pForm, UInt16 controlID) __attribute__ ((section ("printfns")));
static UInt32 getX(FormPtr pForm, UInt16 controlID) __attribute__ ((section ("printfns")));
static void sortControlsOnX(FormPtr pForm, UInt16 **Controls, UInt16 nSize) __attribute__ ((section ("printfns")));
static Boolean isContributionLeft(Line *currentLine) __attribute__ ((section ("printfns")));
static MAB_PRINT_STATUS printLines() __attribute__ ((section ("printfns")));
static MAB_PRINT_STATUS printLine(Line *currentLine) __attribute__ ((section ("printfns")));
static char* getInterControlSpace(Node *thisNode) __attribute__ ((section ("printfns")));
static MAB_PRINT_STATUS sendLineToPrinter(char *szLine) __attribute__ ((section ("printfns")));
static MAB_PRINT_STATUS openPrinterForLinePrint() __attribute__ ((section ("printfns")));
static MAB_PRINT_STATUS closePrinter() __attribute__ ((section ("printfns")));
MAB_PRINT_STATUS MAB_PrintErrorHandler(MAB_PRINT_STATUS status) __attribute__ ((section ("printfns")));
#else
static void getNextLineFromField(Node *thisNode, char **szText);
static void getNextLineFromLabel(Node *thisNode, char **szText);
static void getNextLineFromList(Node *thisNode, char **szText);
static void getNextLineFromNode(Node *thisNode, char **szText);
static Boolean doesNextLineExist(Node *thisItem);
static void initField(Node *thisNode);
static void initLabel(Node *thisNode);
static void initList(Node *thisNode);
static Node* initItemNode(FormPtr pForm, UInt32 controlID);
static UInt32 getCharsVisible(FontID font,UInt16 width);
static void getLineFromBuffer(UInt32 nLine, UInt32 visibleChars, CharPtr szBuffer, char **szText);
static void getLineFromList(FormPtr pForm,UInt16 listID, UInt32 lineNo, CharPtr szListData, char **szLine);
static void createLines(FormPtr pForm, UInt16 ControlsOnForm[], UInt16 nSize);
static Line* initLine(UInt32 nLineNo);
static void sortControls(FormPtr pForm, UInt16 **Controls, UInt16 nSize);
static void createPrintableForm(FormPtr pForm, UInt16 ControlsOnForm[], UInt16 nSize);
static UInt32 getY(FormPtr pForm, UInt16 controlID);
static UInt32 getX(FormPtr pForm, UInt16 controlID);
static void sortControlsOnX(FormPtr pForm, UInt16 **Controls, UInt16 nSize);
static Boolean isContributionLeft(Line *currentLine);
static MAB_PRINT_STATUS printLines();
static MAB_PRINT_STATUS printLine(Line *currentLine);
static char* getInterControlSpace(Node *thisNode);
static MAB_PRINT_STATUS sendLineToPrinter(char *szLine);
static MAB_PRINT_STATUS openPrinterForLinePrint();
static MAB_PRINT_STATUS closePrinter();
MAB_PRINT_STATUS MAB_PrintErrorHandler(MAB_PRINT_STATUS status);
#endif

MAB_PRINT_STATUS MAB_Print_ClosePrinter(MAB_PrintLibrary lib){
	return closePrinter();
}

static MAB_PRINT_STATUS closePrinter(){
	UInt32 result=0;
	LocalID	dbID=DmFindDatabase(0,"PalmPrint");
	Err err;
	if(dbID){
		err = SysAppLaunch(0,dbID,0,cmdEndPrint,"",&result);
	}
	return dbID!= 0 & result == 0?MAB_PRINT_SUCCESS_CODE:MAB_PRINT_ERROR_UNKNOWN_CODE;
}

MAB_PRINT_STATUS MAB_Print_InitPrinter(MAB_PrintLibrary lib){
	return openPrinterForLinePrint();
	
}

static MAB_PRINT_STATUS openPrinterForLinePrint(){
	UInt32 result=0;
	LocalID	dbID=DmFindDatabase(0,"PalmPrint");
	Err err;
	if(dbID){
		err = SysAppLaunch(0,dbID,0,cmdStartPrint,"",&result);
	}
	else
		return MAB_PRINT_INIT_FAILED_CODE;
	return dbID!=0 & result == 0?MAB_PRINT_SUCCESS_CODE:MAB_PRINT_ERROR_UNKNOWN_CODE;
}

MAB_PRINT_STATUS MAB_Print_Line(char *buffer){
	if(buffer)
		return sendLineToPrinter(buffer);
	
	return MAB_PRINT_SUCCESS_CODE;
}

static MAB_PRINT_STATUS sendLineToPrinter(char *szLine){
	MAB_PRINT_STATUS status = MAB_PRINT_SUCCESS_CODE;
	LocalID	dbID=DmFindDatabase(0,"PalmPrint");
	UInt32 result = 0;
	Err err = 0;
	if(dbID){
		err = SysAppLaunch(0,dbID,0,cmdPrintLine,szLine,&result);
		if(result!=0){
			SysAppLaunch(0,dbID,0,cmdEndPrint,szLine,&result);
			return (status = MAB_PRINT_ERROR_UNKNOWN_CODE);
		}
	}
	return status;
}


static MAB_PRINT_STATUS printLines(){
	Line *currentLine;
	MAB_PRINT_STATUS status;
	if((status = openPrinterForLinePrint()) == MAB_PRINT_SUCCESS_CODE){
		currentLine = gFirstLine;
		while(currentLine){
			printLine(currentLine);
			currentLine = currentLine->nextLine;
		}
		status = closePrinter();	
	}
	return status;
}

static MAB_PRINT_STATUS printLine(Line *currentLine){
	char *szLine = NULL;
	char *szNodeText;
	char *szTemp;
	Node *currentNode = NULL;
	Boolean bContinue = true;
	MAB_PRINT_STATUS status = MAB_PRINT_SUCCESS_CODE;
	
	currentNode = currentLine->first;
	while(isContributionLeft(currentLine)){
		getNextLineFromNode(currentNode, &szNodeText);
		szTemp = szLine;
		if(szTemp)
			szLine = MemPtrNew(StrLen(szTemp) + StrLen(szNodeText) + 2);
		else //First time
			szLine = MemPtrNew(StrLen(szNodeText) + 1);
		if(szLine){
			if(szTemp){
				StrCopy(szLine,szTemp);
				StrCat(szLine,szNodeText);
				MemPtrFree(szTemp);
				if(szNodeText)
					MemPtrFree(szNodeText);
			}
			else{
				StrCopy(szLine,szNodeText);
				if(szNodeText)
					MemPtrFree(szNodeText);
			}
		}

		currentNode = currentNode->next;
		if(currentNode == NULL){
//			FrmCustomAlert(900,szLine,"",""); //send this line to printing device.
			status = sendLineToPrinter(szLine);
			if(status != MAB_PRINT_SUCCESS_CODE)
				return status;
			currentNode = currentLine->first;
			if(szLine){
				MemPtrFree(szLine);
				szLine = NULL;
			}
		}
	}
	if(szLine){
//		FrmCustomAlert(900,szLine,"",""); //send this line to printing device.
		status = sendLineToPrinter(szLine);
		MemPtrFree(szLine);
		szLine = NULL;
	}
	
	return status;
}

static Boolean isContributionLeft(Line *currentLine){
	Node *currentNode = NULL;
	Boolean bContribution = false;
	currentNode = currentLine->first;
	while(currentNode){
		bContribution = doesNextLineExist(currentNode);
		if(bContribution == true)
			break;
		currentNode = currentNode->next;
	}
	return bContribution;
}

static void getNextLineFromNode(Node *thisNode, char **szText){
	char *szLine;
	char *spaces = NULL;
	char *szTemp;
	FormObjectKind kind = FrmGetObjectType(thisNode->pForm, FrmGetObjectIndex(thisNode->pForm,thisNode->controlID));
	switch(kind){
		case frmFieldObj:
			getNextLineFromField(thisNode, &szLine);
			break;
		case frmLabelObj:
			getNextLineFromLabel(thisNode, &szLine);
			break;
		case frmListObj:
			getNextLineFromList(thisNode, &szLine);
			break;
	}

	spaces = getInterControlSpace(thisNode);
	if(spaces){
		szTemp = szLine;
		szLine = MemPtrNew(StrLen(szTemp) + StrLen(spaces) + 1);
		StrCopy(szLine,szTemp);
		StrCat(szLine,spaces);
		if(szTemp && szTemp[0] != '\0')
			MemPtrFree(szTemp);
	}
	*szText = szLine;
}


static char* getInterControlSpace(Node *thisNode){
	char *spaces;
	RectangleType c1 ,c2;
	Node *nextNode;
	UInt32 width=0,charsVisible;
	if(thisNode->next != NULL){
		FrmGetObjectBounds(thisNode->pForm,FrmGetObjectIndex(thisNode->pForm,thisNode->controlID),&c1);
		nextNode = thisNode->next;
		FrmGetObjectBounds(nextNode->pForm,FrmGetObjectIndex(nextNode->pForm,nextNode->controlID),&c2);
		width = c2.topLeft.x - (c1.extent.x + c1.topLeft.x);
	}
	charsVisible = getCharsVisible(stdFont,width) + 2; //Add 2 to make sure that there is atleast one space between them.
	spaces = (char *)MemPtrNew(charsVisible +  1);
	MemSet(spaces,charsVisible,' ');
	spaces[charsVisible] = '\0';
	return spaces;
}

static void getNextLineFromField(Node *thisNode, char **szText){
	char *szLine = NULL;
	UInt32 visibleCh;
	RectangleType bounds;
	FontID fnt = FldGetFont(FrmGetObjectPtr(thisNode->pForm,FrmGetObjectIndex(thisNode->pForm,thisNode->controlID)));
	FrmGetObjectBounds(thisNode->pForm,FrmGetObjectIndex(thisNode->pForm,thisNode->controlID),&bounds);

	if(thisNode->nCurrentLine < thisNode->nMaxLines){
		if(thisNode->data){
			szLine = (char *)MemPtrNew(StrLen(thisNode->data)+1);
			StrCopy(szLine, thisNode->data);
		}
		thisNode->nCurrentLine++;
	}
	else{
		visibleCh = getCharsVisible(fnt,bounds.extent.x);
		szLine =  (char *)MemPtrNew(visibleCh + 1);
		MemSet(szLine,visibleCh,' ');
		szLine[visibleCh]='\0';
	}
	*szText = szLine;
}

static void getNextLineFromLabel(Node *thisNode, char **szText){
	char *szLine = NULL;
	UInt32 visibleCh;
	RectangleType bounds;
	FontID fnt = stdFont;
	FrmGetObjectBounds(thisNode->pForm,FrmGetObjectIndex(thisNode->pForm,thisNode->controlID),&bounds);
	
	if(thisNode->nCurrentLine < thisNode->nMaxLines){
		if(thisNode->data){
			szLine = (char *)MemPtrNew(StrLen(thisNode->data)+1);
			StrCopy(szLine, thisNode->data);
		}
		thisNode->nCurrentLine++;
	}
	else{
		visibleCh = getCharsVisible(fnt,bounds.extent.x);
		szLine =  (char *)MemPtrNew(visibleCh + 1);
		MemSet(szLine,visibleCh,' ');
		szLine[visibleCh]='\0';
	}
	*szText = szLine;

}

static void getNextLineFromList(Node *thisNode, char **szText){
	char *szLine = NULL;
	UInt32 visibleCh;
	RectangleType bounds;
	FontID fnt = stdFont;
	FrmGetObjectBounds(thisNode->pForm,FrmGetObjectIndex(thisNode->pForm,thisNode->controlID),&bounds);

	if(thisNode->nCurrentLine < thisNode->nMaxLines){
		szLine = (CharPtr)MemPtrNew(CHARSPERROW + 1);
		MemSet(szLine,CHARSPERROW + 1,0);
		getLineFromList(thisNode->pForm,thisNode->controlID,thisNode->nCurrentLine++,thisNode->data,&szLine);
	}
	else{
		visibleCh = getCharsVisible(fnt,bounds.extent.x);
		szLine =  (char *)MemPtrNew(visibleCh + 1);
		MemSet(szLine,visibleCh,' ');
		szLine[visibleCh]='\0';
	}
	*szText = szLine;
}


static Boolean doesNextLineExist(Node *thisItem){
	return thisItem->nCurrentLine < thisItem->nMaxLines ? true : false;
}


static Node* initItemNode(FormPtr pForm, UInt32 controlID){
	Node *thisNode = NULL;
	FormObjectKind kind;
	
	thisNode = (Node *)MemPtrNew(sizeof(Node));
	if(thisNode){
		thisNode->pForm 		= pForm;
		thisNode->controlID 	= controlID;
		thisNode->next 			= NULL;
		thisNode->nCurrentLine  = 0;
		kind = FrmGetObjectType(thisNode->pForm, FrmGetObjectIndex(thisNode->pForm,thisNode->controlID));
		switch(kind){
			case frmFieldObj:
				initField(thisNode);
				break;
			case frmLabelObj:
				initLabel(thisNode);
				break;
			case frmListObj:
				initList(thisNode);
				break;
		}	

	}
	return thisNode;
}


static void initField(Node *thisNode){
	char *szText = NULL;
	MemHandle h;
	FieldPtr pFld = FrmGetObjectPtr(thisNode->pForm,FrmGetObjectIndex(thisNode->pForm,thisNode->controlID));
	RectangleType bounds;
	FontID fnt = FldGetFont(pFld);
	UInt32 visibleCh; 
	FldGetBounds(pFld,&bounds);
	visibleCh = getCharsVisible(fnt,bounds.extent.x);
	
	h = FldGetTextHandle(pFld);
	if(h){
		szText = (char *)MemHandleLock(h);
		if(szText && szText[0] != '\0'){
			thisNode->data = (char *)MemPtrNew(visibleCh + 1);
			MemSet(thisNode->data,visibleCh,' ');
			thisNode->data[visibleCh] = '\0';
			if(StrLen(szText) > visibleCh)
				StrNCopy(thisNode->data,szText,visibleCh);
			else
				MemMove(thisNode->data,szText,StrLen(szText) + 1);
		}
		else{
			thisNode->data = (char *)MemPtrNew(visibleCh + 1);
			MemSet(thisNode->data,visibleCh,' ');
			thisNode->data[visibleCh] = '\0';
		}
		MemHandleUnlock(h);
	}
	thisNode->nMaxLines = 1;
}


static void initLabel(Node *thisNode){
	char *szText = NULL;
	ControlType *pLabel = FrmGetObjectPtr(thisNode->pForm,FrmGetObjectIndex(thisNode->pForm,thisNode->controlID));
	RectangleType bounds;
	FontID fnt = stdFont;
	UInt32 visibleCh; 
	FrmGetObjectBounds(thisNode->pForm,FrmGetObjectIndex(thisNode->pForm,thisNode->controlID),&bounds);
	visibleCh = getCharsVisible(fnt,bounds.extent.x) + 1;
	
	szText = (char *)CtlGetLabel(pLabel);
	if(szText && szText[0] != '\0'){
		thisNode->data = (char *)MemPtrNew(visibleCh + 1);
		MemSet(thisNode->data,visibleCh,' ');
		thisNode->data[visibleCh] = '\0';
		if(StrLen(szText) > visibleCh)
			StrNCopy(thisNode->data,szText,visibleCh);
		else
			MemMove(thisNode->data,szText,StrLen(szText)+1);
	}
	else{
			thisNode->data = (char *)MemPtrNew(visibleCh + 1);
			MemSet(thisNode->data,visibleCh,' ');
			thisNode->data[visibleCh] = '\0';
	}

	thisNode->nMaxLines = 1;
}


static void initList(Node *thisNode){
	if(gFunc)
		(*gFunc)(thisNode->pForm,thisNode->controlID,&thisNode->data,&thisNode->nMaxLines);
	else{
		thisNode->data = NULL;
		thisNode->nMaxLines = 1;
	}
}


static UInt32 getCharsVisible(FontID font,UInt16 width){
	UInt32 CharsVisiblePerLine = CHARSPERROW;
//	UInt32 temp = CharsVisiblePerLine;
	double delta=0;
	switch(font){
		case stdFont:
			CharsVisiblePerLine = width/5;
			break;
		case boldFont:
			CharsVisiblePerLine = width/6;
			break;
		case largeFont:
			CharsVisiblePerLine = width/7;
			break;
	}
	return CharsVisiblePerLine;
}


static void getLineFromBuffer(UInt32 nLine, UInt32 visibleChars, CharPtr szBuffer, char **szText){
	char chPad = ' ';
	CharPtr szBuf = szBuffer + nLine*(CHARSPERROW+1);
	int i=0;
	for(i=0; i<visibleChars; i++){
		if(szBuf[i] != '\0'){
			*((*szText) + i) = szBuf[i];
		}
		else{
			*((*szText) + i) = chPad;
		}
	}
	
}

static void getLineFromList(FormPtr pForm,UInt16 listID, UInt32 lineNo, CharPtr szListData, char **szLine){
	ListPtr pList;

	pList = (ListPtr)FrmGetObjectPtr(pForm,FrmGetObjectIndex(pForm,listID));
	getLineFromBuffer(lineNo,getCharsVisible(pList->font,pList->bounds.extent.x),szListData,szLine);
}

static void createLines(FormPtr pForm, UInt16 ControlsOnForm[], UInt16 nSize){
	int cnt=0;
	sortControls(pForm, &ControlsOnForm, nSize);
	createPrintableForm(pForm,ControlsOnForm,nSize);
}


static void createPrintableForm(FormPtr pForm, UInt16 ControlsOnForm[], UInt16 nSize){
	int i=0;
	int lastY = getY(pForm,ControlsOnForm[0]) - TOLERANCE;
	signed int delta;
	UInt32 nLineCnt = 0;
	Line *thisLine, *newLine;
	Node *currentNode, *nextNode;
	
	for(i=0; i<nSize; i++){
		delta = getY(pForm,ControlsOnForm[i]) - lastY;
		if(delta < TOLERANCE){//add to existing line
			nextNode = initItemNode(pForm, ControlsOnForm[i]);
			currentNode->next = nextNode;
			thisLine->nElements++;
			currentNode = currentNode->next;
		}
		else{//create new line and this element to the line.
			newLine = initLine(nLineCnt);
			if(newLine == NULL){
				//error message
				break;
			}
			else{
				if(nLineCnt==0){
					gFirstLine = newLine;
					thisLine = newLine;
				}
				else{
					thisLine->nextLine = newLine;
					thisLine = thisLine->nextLine;
				}
				currentNode = initItemNode(pForm, ControlsOnForm[i]);
				if(currentNode == NULL){
					//error messgae
					break;
				}
				else{
					thisLine->first = currentNode;
					thisLine->nElements++; 
					nLineCnt++;
				}
			}
		}
		lastY = getY(pForm,ControlsOnForm[i]);
	}
}

static Line* initLine(UInt32 nLineNo){
	Line *line = NULL;
	line = (Line *)MemPtrNew(sizeof(Line));
	if(line){
		line->nLineNo = nLineNo;
		line->nextLine = NULL;
		line->nElements = 0;
	}
	return line;
}


static void sortControls(FormPtr pForm, UInt16 **Controls, UInt16 nSize){
	UInt16 *controls = *Controls;
	int i=0, j=0;
	UInt16 temp;
	
	for(i=0; i<nSize; i++){
		for(j=i; j<nSize; j++){
			if(getY(pForm,controls[j]) < getY(pForm,controls[i])){
				temp 		= controls[j];
				controls[j] = controls[i];
				controls[i] = temp;
			}
		}
	}
	sortControlsOnX(pForm, &controls, nSize);
}

static void sortControlsOnX(FormPtr pForm, UInt16 **Controls, UInt16 nSize){
	UInt16 *controls = *Controls;
	int i=0, j=0;
	UInt16 temp;
	signed int delta;
	
	for(i=0; i<nSize; i++){
		for(j=i; j<nSize; j++){
			delta = (getY(pForm,controls[j]) - getY(pForm,controls[i]));
			if( delta < TOLERANCE && getX(pForm,controls[j]) < getX(pForm,controls[i])){
				temp 		= controls[j];
				controls[j] = controls[i];
				controls[i] = temp;
			}
		}
	}
}

static UInt32 getY(FormPtr pForm, UInt16 controlID){
	RectangleType pBounds;
	FrmGetObjectBounds(pForm,FrmGetObjectIndex(pForm,controlID),&pBounds);
	return pBounds.topLeft.y;
}

static UInt32 getX(FormPtr pForm, UInt16 controlID){
	RectangleType pBounds;
	FrmGetObjectBounds(pForm,FrmGetObjectIndex(pForm,controlID),&pBounds);
	return pBounds.topLeft.x;
}

static void freeAll(){
	Line *thisLine = gFirstLine;
	Line *nextLine;
	Node *currentNode;
	int i;
	
	for(thisLine=gFirstLine; thisLine!=NULL;){
		Node **nodeStack = (Node**) MemPtrNew(sizeof(Node)*thisLine->nElements);
		for(currentNode = thisLine->first, i=0; currentNode != NULL; currentNode=currentNode->next,i++){
			nodeStack[i] = currentNode;
		}
		for(i=0; i<thisLine->nElements; i++){
			FormObjectKind kind;
			currentNode = nodeStack[i];
			if(currentNode != NULL){
				kind = FrmGetObjectType(currentNode->pForm, FrmGetObjectIndex(currentNode->pForm,currentNode->controlID));
				switch(kind){
					case frmFieldObj:
						if(currentNode->data != NULL)
							MemPtrFree(currentNode->data);
						break;
					case frmLabelObj:
						if(currentNode->data != NULL)
							MemPtrFree(currentNode->data);
						break;
					case frmListObj:
						// we do not free the data associated with list.
						break;
				}
				MemPtrFree(currentNode); // now free the node itself.
				currentNode = NULL;
			}
		}
		MemPtrFree(nodeStack);
		
		
		nextLine = thisLine->nextLine;
		if(thisLine)
			MemPtrFree(thisLine);
		thisLine = nextLine;
	}
	if(nextLine)
		MemPtrFree(nextLine);
}

MAB_PRINT_STATUS MAB_Print_PrintForm(MAB_PrintLibrary lib, FormPtr pForm, MAB_Print_ListDataHandlerPtr func){
	MAB_PRINT_STATUS status = MAB_PRINT_SUCCESS_CODE;
	
	UInt16 *ControlsOnForm = (UInt16 *)MemPtrNew(sizeof(UInt16)*pForm->numObjects);
	FieldType *field;
	ControlType *label;
	ListType *list;
	int i;
	UInt16 nSize=0;
	Int16 val;
#if __COMMENTED_CODE__
	char szSize[10];
	UInt32 ramsize;
	UInt32 freebytes;
	MemCardInfo(0,0,0,0,0,0, &ramsize,&freebytes);
	
	StrIToA(szSize,ramsize);
	FrmCustomAlert(900,szSize,"RAM size","");
	StrIToA(szSize,freebytes);
	FrmCustomAlert(900,szSize,"free bytes","");
#endif
	if(!ControlsOnForm)
		return MAB_PRINT_ERROR_UNKNOWN_CODE;

	for(i=0; i<pForm->numObjects; i++){
		FormObjListType obj = pForm->objects[i];
		FormObjectKind kind = obj.objectType;
		switch(kind){
			case frmListObj:
				list = (ListType*)obj.object.list;
				if(list->attr.usable == true)
					ControlsOnForm[nSize++] = list->id;
				break;
			case frmFieldObj:
				field = (FieldType*)obj.object.field;
				if(field->attr.usable == true)
					ControlsOnForm[nSize++] = field->id;
				break;
				
			case frmLabelObj:
				label = (ControlType*)obj.object.control;
				val = CtlGetValue(label);
				if(label->attr.enabled == true)
					ControlsOnForm[nSize++] = label->id;
				break;
		}
	}  	
	gFunc = func;	
	createLines(pForm, ControlsOnForm, nSize);
	MAB_PrintErrorHandler(status = printLines());
	freeAll();
	if(ControlsOnForm)
		MemPtrFree(ControlsOnForm);
	
#if __COMMENTED_CODE__
	MemCardInfo(0,0,0,0,0,0, &ramsize,&freebytes);
	StrIToA(szSize,freebytes);
	FrmCustomAlert(900,szSize,"free bytes","");
#endif
	
	return status;
}

MAB_PRINT_STATUS MAB_Print_Buffer(MAB_PrintLibrary lib, char *buffer){
    DWord    result;
    LocalID  dbID;
    Err err;
    if(buffer == NULL)
    	return MAB_PRINT_SUCCESS_CODE;
    dbID=DmFindDatabase(0,"PalmPrint");
    if (dbID) {
       err=SysAppLaunch(0,dbID,0,cmdPrintChars,buffer,&result);
    }
    if(result != 0)
	    MAB_PrintErrorHandler(MAB_PRINT_ERROR_UNKNOWN_CODE);
	else
		MAB_PrintErrorHandler(MAB_PRINT_SUCCESS_CODE);
	return(( dbID!= 0 & result == 0)?MAB_PRINT_SUCCESS_CODE:MAB_PRINT_ERROR_UNKNOWN_CODE);
}

MAB_PRINT_STATUS MAB_PrintErrorHandler(MAB_PRINT_STATUS status){
	switch(status){
		case MAB_PRINT_INIT_FAILED_CODE:
		case MAB_PRINT_ERROR_UNKNOWN_CODE:
		case MAB_PRINT_COMMUNICATIONS_ERROR_CODE:
			FrmCustomAlert(MAB_PRINT_ALERT_ID,MAB_PRINT_PRINT_FAILED,"","");
			break;
			
		case MAB_PRINT_SUCCESS_CODE:
			FrmCustomAlert(MAB_PRINT_ALERT_ID,MAB_PRINT_SUCCESS,"","");
			break;
	}
	return status;
}


void MAB_Print_SetAlertIDForMessages(UInt16 alertID){
	MAB_PRINT_ALERT_ID = alertID;
}
