/*****************************************************************************
 **
 ** Licensed Material - Property of IBM
 **
 ** 5724-D04
 **
 ** (C) Copyright IBM Corp. 2000 - 2004  All Rights Reserved
 **
 ** US Government Users Restricted Rights - Use, duplication or
 ** disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 **
 ***************************************************************************/
/**
* File 				:	MAB_Scanner.h
*
**/

#include "ScanMgrStruct.h"
#include "ScanMgrDef.h"
#include "ScanMgr.h"

#define COUNT_SUPPORTED_BARCODES 19 // number of barcodes supported by MAB.
#define MAB_SCAN_STATUS UInt16

#define MAB_SCAN_STATUS_OK 				STATUS_OK
#define	MAB_SCAN_UNKNOWN_ERROR 			ERROR_UNDEFINED
#define	MAB_SCAN_BARCODE_NOT_SUPPORTED  NOT_SUPPORTED
#define	MAB_SCAN_COMMUNICATIONS_ERROR	COMMUNICATIONS_ERROR
#define	MAB_SCAN_BAD_PARAM				BAD_PARAM
#define	MAB_SCAN_BATCH_ERROR			BATCH_ERROR
#define	MAB_SCAN_NODECODE				1001
#define	MAB_SCAN_SOFTSCAN_FAILED		1002
#define MAB_SCAN_CURRENT_FIELD_NOT_SCANAWARE 1003

//Datastructure used by error Handler.
typedef struct ErrorHandlerStruct{
	UInt16 alertID; //Alert to be used to display the message.
	UInt16 currentStatus;	//Error Code.
	CharPtr szErrorMsg; //Error Message.
}ErrorHandlerData;
//Global Initialization.
static struct ErrorHandlerStruct gDefaultScanErrorHandlerData = {903,MAB_SCAN_STATUS_OK,NULL};

//Datastructures used by initializer.
static UInt16 supportedBarcodes[] = {
									barCODE39,
									barUPCA,
									barUPCE,
									barUPCE1,
									barEAN13,
									barEAN8,
									barD25,
									barI2OF5,
									barCODABAR,
									barCODE128,
									barCODE93,
									barTRIOPTIC39,
									barUCC_EAN128,
									barMSI_PLESSEY,
									barUPCE1,
									barBOOKLAND_EAN,
									barISBT128,
									barCOUPON,
									barPDF417
									};

static Boolean isBarCodeEnabled[] =	{
									true,
									true,
									true,
									true,
									true,
									true,
									true,
									true,
									true,
									true,
									true,
									true,
									true,
									true,
									true,
									true,
									true,
									true,
									true
								};

typedef struct ScannerInitParmsStruct{
	UInt16 triggeringMode; 									//allowed triggering mode LEVEL,HOST
	UInt16 aimDuration;										//when triggering mode HOST this value must be set or else a default will be set.
	UInt16 *barCodeTypes;									//Barcodes supported by MAB
	Boolean *barCodesToEnable;		//This value will be used as the 2nd param by ScanSetBarcodeEnabled
	Boolean beepAfterGoodDecode;
	UInt16  DecodeFrequency;
}ScannerInitParms;

static struct ScannerInitParmsStruct gDefault_HardScan_ScannerInitData = {HOST,20,supportedBarcodes,isBarCodeEnabled,true,3000};

#define _SCAN_STATUS gDefaultScanErrorHandlerData.currentStatus
#define _SCAN_MESSAGE gDefaultScanErrorHandlerData.szErrorMsg
#define _HARDSCAN gDefault_HardScan_ScannerInitData
#define _MAB_SCAN_TRIGGERING_MODE gDefault_HardScan_ScannerInitData.triggeringMode
#define _MAB_SCAN_AIM_DURATION gDefault_HardScan_ScannerInitData.aimDuration


#if __MULTI_SEG_APP__

MAB_SCAN_STATUS MAB_InitializeScanner() __attribute__ ((section ("scanfns")));
MAB_SCAN_STATUS MAB_DeInitScanner(void) __attribute__ ((section ("scanfns")));
CharPtr MAB_DecodeScan (void) __attribute__ ((section ("scanfns")));
CharPtr DecodeBarCode(void) __attribute__ ((section ("scanfns")));
Boolean MAB_ScanErrorHandler(UInt16 status) __attribute__ ((section ("scanfns")));
MAB_SCAN_STATUS InitForHardScan() __attribute__ ((section ("scanfns")));
Boolean AssertScanStatus(Boolean bAssert,UInt16 nLineNo) __attribute__ ((section ("scanfns")));
MAB_SCAN_STATUS MAB_DoSoftScan() __attribute__ ((section ("scanfns")));
Word MAB_ScanIsFocusOnScanAwareField(FormPtr pFrm, Word scanAwareFields[], UInt16 nCount) __attribute__ ((section ("scanfns")));
void MAB_ScanSetAlertIDForMessages(Word alertID) __attribute__ ((section ("scanfns")));
void MAB_ScanHandleNoDecode() __attribute__ ((section ("scanfns")));
Word MAB_ScanGetNextScanAwareField(FormPtr pFrm, Word lastField, Word scanAwareFields[], UInt16 nCount) __attribute__ ((section ("scanfns")));
void MAB_ScanDisableBarCodeType(Word barCodeType) __attribute__ ((section ("scanfns")));
void MAB_ScanEnableBarCodeType(Word barCodeType)  __attribute__ ((section ("scanfns")));
Boolean MAB_ScanSetAimDuration(UInt16 duration)   __attribute__ ((section ("scanfns")));
Boolean MAB_ScanSetTriggeringMode(UInt16 mode)   __attribute__ ((section ("scanfns")));
MAB_SCAN_STATUS MAB_ScanGetLastErrorCode()   __attribute__ ((section ("scanfns")));
CharPtr MAB_ScanGetLastErrorMessage()   __attribute__ ((section ("scanfns")));

#else

MAB_SCAN_STATUS MAB_InitializeScanner(void);
MAB_SCAN_STATUS MAB_DeInitScanner(void);
CharPtr MAB_DecodeScan (void);
CharPtr DecodeBarCode(void);
Boolean MAB_ScanErrorHandler(UInt16 status);
MAB_SCAN_STATUS InitForHardScan(void);
Boolean AssertScanStatus(Boolean bAssert,UInt16 nLineNo);
MAB_SCAN_STATUS MAB_DoSoftScan(void);
Word MAB_ScanIsFocusOnScanAwareField(FormPtr pFrm, Word scanAwareFields[], UInt16 nCount);
void MAB_ScanSetAlertIDForMessages(Word alertID);
void MAB_ScanHandleNoDecode(void);
Word MAB_ScanGetNextScanAwareField(FormPtr pFrm, Word lastField, Word scanAwareFields[], UInt16 nCount);
void MAB_ScanDisableBarCodeType(Word barCodeType);
void MAB_ScanEnableBarCodeType(Word barCodeType);
Boolean MAB_ScanSetAimDuration(UInt16 duration);
Boolean MAB_ScanSetTriggeringMode(UInt16 mode);
MAB_SCAN_STATUS MAB_ScanGetLastErrorCode();
CharPtr MAB_ScanGetLastErrorMessage();

#endif
