/************************************************************************
* COPYRIGHT:   Copyright    1998 Symbol Technologies, Inc. 
*
* FILE:        ScanMgrStruct.h
*
* SYSTEM:      Symbol barcode scanner for Palm III.
*
* HEADER:      Scan Manager structure definitions
*
* DESCRIPTION: Provides the structure definitions used by the internal
*					functions of the decoder shared library.
*
* HISTORY:     4/13/98    SS   Created
*              ...
*************************************************************************/

#ifndef _ScanMgrStruct_h
#define _ScanMgrStruct_h

#ifdef __cplusplus
   extern "C" {
#endif

#ifndef scrCopy
#include <PalmCompatibility.h>  //fg deleted ?
#endif

// The ScanEvent record.
typedef struct 
{
	//eventsEnum     eType;  //fg deleted ?
	enum events	{eType};   //fg added ?
	Boolean        penDown;
	SWord          screenX;
	SWord          screenY;
	union scanData 
	{
		struct scanGen
		{
			Word data1;		//	Use this variable as flag to tell app if decode data has multiple packets.
			Word data2;		
			Word data3;
			Word data4;
			Word data5;
			Word data6;
			Word data7;
			Word data8;
		} scanGen;

		struct 
		{
			UInt batteryLevel;			// The current voltage measured in millivolts
			UInt batteryErrorType;		// not used
		} batteryError;

	} scanData;	// End of union
	
} ScanEventType;
typedef ScanEventType *ScanEventPtr;

/*******************************************************************
 *    Message structure used to hold decoder messages              *
 * Used by ScanGetDecodeData to return barcode type and data       *
 *******************************************************************/
#define MAX_PACKET_LENGTH       258   //origin for 1D:  258
typedef struct  tagMESSAGE
{
	Int16 length; // length of the data
	Int16 type;	// contains the barcode type when the msg is DecodeData
	Int16 status; // should be STATUS_OK
	unsigned char data[MAX_PACKET_LENGTH]; // the message data
} MESSAGE;

#define MAX_DCD_LENGTH       3000   //origin for 1D:  258
/*******************************************************************
 *                                               *
 *******************************************************************/
#define MAX_DECODER_PARMS       3000   // Origin for 1D:  1024

typedef struct  tagMESSAGE_Dcd
{
	Int16 length; // length of the data
	Int16 type;	// contains the barcode type when the msg is DecodeData
	Int16 status; // should be STATUS_OK
	unsigned char data[MAX_DCD_LENGTH]; // the message data
} MESSAGE_Dcd;

/*******************************************************************
 * BATCH structure used to hold all the decoder parameters to be   *
 * sent to the decoder.  Not used by application programmers.      *
 *******************************************************************/
#define MAX_BATCH_PARAM         247
typedef struct tagBATCH
{
	Int16 length;
	Int16 data[MAX_BATCH_PARAM + 3];
} BATCH;

/*******************************************************************
 * Communication structure COMM_STRUCT is used in the Scan Manager *
 * internal code.                                                  *
 * Note that the communications parameters are preset for this     *
 * version of the decoder and need not be changed by applications. *
 *******************************************************************/
typedef struct  tagCOMM_STRUCT
{
	Int16 port;
	Int16 baud;
	Int16 parity;			
	Int16 stop_bits;
	Int16 host_timeout;
} COMM_STRUCT;


#define PACKET_CONTINUATION	0x0002
#ifdef __cplusplus
    }
#endif

#endif